try:
    # Favor the Py3 module if available
    from collections.abc import Sequence
except ImportError:
    # Fall back to the Py2 module if we have to
    from collections import Sequence
    # And since we now know we're on Py2, get Py3-style super()
    from builtins import super


class NanoDay(object):
    """A pynano historical Day object.

    These objects provide access to historical data for a particular day from
    the history API.
    """
    # Data for this object
    _data = None

    def __init__(self, data):
        """
        :param dict data: A dictionary containing the day's properties
        """
        self._data = data

    @property
    def date(self):
        """The date of this object.

        This property corresponds to `wcdate` or `date` in the API.

        :rtype: string
        """
        return self._data['date']

    @property
    def wordcount(self):
        """The word count on this date.

        This property corresponds to `wc` in the API.

        :rtype: int
        """
        return int(self._data['wc'])


class NanoDaySequence(Sequence):
    _history = None

    def __init__(self, history, *args, **kwargs):
        """Initialize the sequence with our day-to-day history.

        :param dict history: Dictionary (indexed by day-1) of daily data
        """
        super().__init__(*args, **kwargs)
        self._history = history

    def __len__(self):
        """Return the length of our history."""
        return len(self._history)

    def __getitem__(self, key):
        """Retrieve the day represented by the given key.

        The sequence is 0-indexed, so to access data for e.g. the 12th, key
        should be 11.

        :param int key: The day (less 1) to retrieve
        :raises TypeError: if `key` is not an integer
        :raises IndexError: if `key` is out of range
        :returns: Daily data for the object
        :rtype: NanoDay

        .. todo::
           Need to detect the difference between out-of-range index, and a gap
           in the data; if the latter, need an "empty" response.
        """
        try:
            key = int(key)

            return self._history[key]
        except ValueError:
            # Not an integer
            raise TypeError('Index must be an integer: {key}'.format(key=key))
        except KeyError:
            raise IndexError()




import hashlib
import sys


import requests


from .base import NanoBase


if sys.version_info < (3, 0, 0):
    # We're on Py2, get Py3-style super()
    from builtins import super


class User(NanoBase):
    """
    A NanoBase object for representing NaNoWriMo participants.
    """
    # Endpoint URLs for user objects
    _primary_url = 'https://nanowrimo.org/wordcount_api/wc/{name}'
    _history_url = 'https://nanowrimo.org/wordcount_api/wchistory/{name}'
    _writeapi_url = 'https://nanowrimo.org/api/wordcount'

    __secret_key = None

    def __init__(self, name, secret_key=None, **kwargs):
        super().__init__(name, **kwargs)

        self.secret_key = secret_key

    @property
    def id(self):
        """The User's id.

        This property corresponds to `uid` in the API.

        :rtype: string
        """
        return self._fetch_element('uid')

    @property
    def name(self):
        """The User's username.

        This property corresponds to `uname` in the API.

        :rtype: string
        """
        return self._fetch_element('uname')

    @property
    def wordcount(self):
        """The User's current word count.

        This property corresponds to `user_wordcount` in the API.

        :rtype: int
        """
        return int(self._fetch_element('user_wordcount'))

    @wordcount.setter
    def wordcount(self, val):
        """Update the User's current word count.

        This uses the NaNoWriMo WriteAPI and the User's `secret_key` to update
        the wordcount on the NaNoWriMo website.

        :param int val: The User's new word count
        """
        if self.secret_key is None:
            raise ValueError('secret_key must be set before wordcount can be')

        if val != self.wordcount:
            val = int(val)
            name = self.name
            key = self.__secret_key

            h = hashlib.sha1()
            h.update(key.encode())
            h.update(name.encode())
            h.update(str(val).encode())

            data = {
                'hash': h.hexdigest(),
                'name': name,
                'wordcount': val,
            }

            r = requests.put(self._writeapi_url, data=data)

            if not r.ok:
                r.raise_for_status()

    @property
    def winner(self):
        """The User's "winner" status.

        This property corresponds to `winner` in the API.

        :rtype: bool
        """
        return self._fetch_element('winner') == 'true'

    @property
    def secret_key(self):
        """The User's WriteAPI key.

        This is not actually returned; if the key has been set you will get a
        string of '*' characters. Returns None if no key has been set.

        :rtype: string
        """
        if self.__secret_key is not None:
            return '*' * len(self.__secret_key)

        return None

    @secret_key.setter
    def secret_key(self, val):
        """Set the User's WriteAPI key.

        :param string val: The User's WriteAPI secret key
        """
        self.__secret_key = val




from decimal import Decimal


from .base import NanoBase
from .day import NanoDay


class RegionDay(NanoDay):
    """A Day in the history of a region.

    .. note::
       The API does not provide daily word counts for regions, and due to
       the `average` seemingly being based on who posted updates rather than on
       `count`, it cannot at this time be calculated.
    """
    @property
    def wordcount(self):
        """The API doesn't provide daily word counts for regions.

        :raises AttributeError: on all calls due to it not being in the API.
        """
        raise AttributeError("'RegionDay' object has no attribute 'wordcount'")

    @property
    def min(self):
        """Minimum word count posted to the region this day.

        Corresponds to `min` in the API.

        :rtype: int
        """
        return int(self._data['min'])

    @property
    def max(self):
        """Maximum word count posted to the region this day.

        Corresponds to `max` in the API.

        :rtype: int
        """
        return int(self._data['max'])

    @property
    def average(self):
        """Average word count posted to the region this day.

        Corresponds to `average` in the API.

        :rtype: Decimal
        """
        return Decimal(self._data['average'])

    @property
    def stddev(self):
        """Standard deviation of word counts posted to the region this day.

        Corresponds to `stddev` in the API.

        :rtype: Decimal
        """
        return Decimal(self._data['stddev'])

    @property
    def writers(self):
        """Writers in the region.

        Corresponds to `count` in the API.

        :rtype: int
        """
        return int(self._data['count'])

    @property
    def donations(self):
        """Donations made from this region on this day.

        Corresponds to `donations` in the API.

        :rtype: Decimal
        """
        return Decimal(self._data['donations'] or '0.0')

    @property
    def donors(self):
        """Number of people donating from this region on this day.

        Corresponds to `numdonors` in the API.

        :rtype: int
        """
        return int(self._data['numdonors'] or 0)


class Region(NanoBase):
    """A NanoBase object to represent NaNoWriMo regions from the API."""
    # Endpoint URLs for region objects
    _primary_url = 'https://nanowrimo.org/wordcount_api/wcregion/{name}'
    _history_url = 'https://nanowrimo.org/wordcount_api/wcregionhist/{name}'

    # Use our own Day object
    _day_class = RegionDay
    # The day field for the history API
    _date_field = 'date'

    @property
    def id(self):
        """Region ID.

        Corresponds to `rid` in the API.

        :rtype: string
        """
        return self._fetch_element('rid')

    @property
    def name(self):
        """Region name.

        Corresponds to `rname` in the API.

        :rtype: string
        """
        return self._fetch_element('rname')

    @property
    def wordcount(self):
        """Region word count.

        Corresponds to `word count` in the API.

        :rtype: int
        """
        return int(self._fetch_element('region_wordcount'))

    @property
    def min(self):
        """Minimum word count in the region.

        Corresponds to `min` in the API.

        :rtype: int
        """
        return int(self._fetch_element('min'))

    @property
    def max(self):
        """Maximum word count in the region.

        Corresponds to `max` in the API.

        :rtype: int
        """
        return int(self._fetch_element('max'))

    @property
    def average(self):
        """Average word count in the region.

        Corresponds to `average` in the API.

        :rtype: Decimal
        """
        return Decimal(self._fetch_element('average'))

    @property
    def stddev(self):
        """Standard deviation of word counts in the region.

        Corresponds to `stddev` in the API.

        :rtype: Decimal
        """
        return Decimal(self._fetch_element('stddev'))

    @property
    def writers(self):
        """Writers in the region.

        Corresponds to `count` in the API.

        :rtype: int
        """
        return int(self._fetch_element('count', alt_index='numparticipants'))

    @property
    def donations(self):
        """Donations made from the region.

        Corresponds to `donations` in the API.

        :rtype: Decimal
        """
        return Decimal(self._fetch_element('donations'))

    @property
    def donors(self):
        """Donors from the region.

        Corresponds to `numdonors` in the API.

        :rtype: int
        """
        return int(self._fetch_element('numdonors'))




import sys
from decimal import Decimal


from .base import NanoBase
from .day import NanoDay


if sys.version_info < (3, 0, 0):
    # We're on Py2, get Py3-style super()
    from builtins import super


class SiteDay(NanoDay):
    """A Day in the history of the site.
    """
    @property
    def min(self):
        """Minimum word count posted to the region this day.

        Corresponds to `min` in the API.

        :rtype: int
        """
        return int(self._data['min'])

    @property
    def max(self):
        """Maximum word count posted to the region this day.

        Corresponds to `max` in the API.

        :rtype: int
        """
        return int(self._data['max'])

    @property
    def average(self):
        """Average of word counts posted to the region this day.

        Corresponds to `average` in the API.

        :rtype: Decimal
        """
        return Decimal(self._data['average'])

    @property
    def stddev(self):
        """Standard deviation of word counts posted to the region this day.

        Corresponds to `stddev` in the API.

        :rtype: Decimal
        """
        return Decimal(self._data['stddev'])

    @property
    def count(self):
        """Corresponds to `count` in the API

        .. todo::
           Figure out what this one actually is.

        :rtype: int
        """
        return int(self._data['count'])


class Site(NanoBase):
    # Endpoint URLs for region objects
    _primary_url = 'https://nanowrimo.org/wordcount_api/wcstatssummary{name}'
    _history_url = 'https://nanowrimo.org/wordcount_api/wcstats{name}'

    # Use our own Day object
    _day_class = SiteDay

    def __init__(self, *args, **kwargs):
        """
        :param bool prefetch: If True, the API will be queried immediately;
                              otherwise it is queried only when needed

        .. note::
           Site objects, unlike other NaNoWriMo API objects, do not have a name
           attribute. We override the __init__() method to remove the name
           parameter and supply an empty string to the parent's method.
        """
        super().__init__(name='', *args, **kwargs)

    @property
    def wordcount(self):
        """Word count for the site.

        Corresponds to `site_wordcount` in the API.

        :rtype: int
        """
        # History API has alternative names for some reason, try those too
        wordcount = self._fetch_element('site_wordcount',
                                        alt_index='site_word_count')

        return int(wordcount)

    @property
    def min(self):
        """Minimum word count for the site.

        Corresponds to `min` in the API.

        :rtype: int
        """
        return int(self._fetch_element('min'))

    @property
    def max(self):
        """Maximum word count for the site.

        Corresponds to `max` in the API.

        :rtype: int
        """
        return int(self._fetch_element('max'))

    @property
    def average(self):
        """Average word count for the site.

        Corresponds to `average` in the API.

        :rtype: Decimal
        """
        return Decimal(self._fetch_element('average'))

    @property
    def stddev(self):
        """Standard deviation of word counts for the site.

        Corresponds to `stddev` in the API.

        :rtype: Decimal
        """
        return Decimal(self._fetch_element('stddev'))

    @property
    def writers(self):
        """Number of writers participating on the site.

        Corresponds to `count` in the API.

        :rtype: int
        """
        # History API has alternative names for some reason, try those too
        writers = self._fetch_element('count', alt_index='numparticipants')

        return int(writers)




"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

import sys
from codecs import open  # To use a consistent encoding
from os import path
from setuptools import find_packages, setup  # Prefer setuptools over distutils

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# Our package dependencies
deps = [
    'requests>=2.9.1',
    'xmltodict>=0.9.2',
]
if sys.version_info < (3, 0, 0):
    # The future package is a compatibility layer between our Py3 code and Py2
    deps += ['future>=0.15.2']

setup(
    name='pynano',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='1.0.1',

    description='A simple Python library for the NaNoWriMo API',
    long_description=long_description,

    # The project's main homepage.
    url='https://pynano.readthedocs.org/',

    # Author details
    author='Travis Veazey',
    author_email='travisvz@gmail.com',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Intended Audience :: Developers',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',

        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],

    # What does your project relate to?
    keywords='nanowrimo wordcount api',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=deps,

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
            'test': [
                # Some of these don't work from tests_require for some reason;
                # putting all of them here for consistency.
                'coverage>=4.0.3',
                'docutils>=0.12',
                'flake8>=2.5.1',
                'flake8-import-order>=0.6.1',
                'pep8-naming>=0.3.3',
                'responses>=0.5.0',
                ]
    },

    # Integrate `setup.py test` with pytest
    setup_requires=[
            'pytest-runner',
            ],
    tests_require=[
            'pytest',
            ],

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
    },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    data_files=[],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
    },
)



import requests
import xmltodict


from .day import NanoDay, NanoDaySequence


class NanoBase(object):
    """Base object for pynano API objects.

    This object implements the common functionality for fetching and processing
    API data from the NaNoWriMo API. By default the API is queried "lazily",
    that is requests are only sent when data API is actually requested. This
    behavior can be overridden with the *prefetch* parameter.
    """

    # API name for this object
    _name = None

    # Endpoint URLs for this object; must include a '{name}' placeholder
    _primary_url = None
    _history_url = None

    # The history day class for this object
    _day_class = NanoDay
    # The API field with a given day's date
    _date_field = 'wcdate'

    # Caches for retrieved data
    _data = None
    _history = None

    def __init__(self, name, prefetch=False):
        """
        :param str name: The name of the current object in the NaNoWriMo API
        :param bool prefetch: If True, the API will be queried immediately;
                              otherwise it is queried only when needed
        """

        # Pre-process name before we save it
        name = name.replace(' :: ', '-').replace(' ', '-')
        self._name = name

        if prefetch:
            self._fetch()

    @property
    def history(self):
        """Historical data for this object.

        :rtype: NanoDaySequence
        """
        if self._history is None:
            # Haven't fetched history data yet, do so now
            self._fetch(True)

        # Return the history object
        return self._history

    def _fetch_element(self, index, alt_index=None):
        """Get a particular data element, querying the API if necessary."""
        # We try from alt_index first
        my_index = alt_index or index

        try:
            # Attempt to return the requested data element
            return self._data[my_index]
        except (KeyError, TypeError):
            # Didn't find it, or nothing fetched yet
            if alt_index is not None:
                # Retry without alt_index
                return self._fetch_element(index)
            else:
                # Fetch data from the API
                self._fetch()

            # If we still don't find it, allow the exception to be raised
            return self._data[my_index]

    def _fetch(self, history=False):
        """Fetch data from the API.

        If the history parameter is True, then the history URL will be queried
        for data; if it is False, then the primary URL will be queried.
        """
        if history:
            # Fetching history data
            url = self._history_url
        else:
            # Fetching primary data
            url = self._primary_url

        # Now fetch from the server
        r = requests.get(url.format(name=self._name))
        # Parse the returned data (removing root element)
        data = next(iter(xmltodict.parse(r.text).values()))

        # Now stash the data
        if history:
            # Stash the history data
            self._history = self._process_history(data['wordcounts'])

            if self._data is None:
                # Haven't fetched primary data, stash what we got
                del data['wordcounts']
                self._data = data
        else:
            # Stash the primary data
            self._data = data

    def _process_history(self, history_data):
        """Process the history API's data into an easily-indexed dict."""
        processed = {}

        for data in history_data['wcentry']:
            # Get the date from the proper field
            date = data[self._date_field]

            # Normalize the date field
            del data[self._date_field]
            data['date'] = date

            # Break out the date elements
            year, month, day = date.split('-')

            if month != '11':
                # Do nothing if we somehow got non-November data
                continue

            # Construct a Day object from our data
            the_day = self._day_class(data)
            # Index the data by day less 1 (0-indexed data)
            processed[int(day) - 1] = the_day

        return NanoDaySequence(processed)

